@@ -1,5 +1,5 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003 Adam Sampson <azz@us-lot.org>
+# Copyright 2003, 2004 Adam Sampson <azz@us-lot.org>
 #
 # rawdog is free software; you can redistribute and/or modify it
 # under the terms of that license as published by the Free Software
@@ -16,10 +16,12 @@
 # Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 # MA 02111-1307 USA, or see http://www.gnu.org/.
 
-VERSION = "1.13"
-import feedparser
+VERSION = "2.4"
+STATE_VERSION = 2
+import feedparser, feedfinder, plugins
 from persister import Persistable, Persister
-import os, time, sha, getopt, sys, re, urlparse, cgi, socket
+import os, time, sha, getopt, sys, re, urlparse, cgi, socket, urllib2, calendar
+import string
 from StringIO import StringIO
 
 def set_socket_timeout(n):
@@ -39,23 +41,6 @@ def format_time(secs, config):
 		format = config["timeformat"] + ", " + config["dayformat"]
 	return time.strftime(format, t)
 
-def select_content(contents):
-	"""Given a list of contents with alternative content types, select a
-	preferred version."""
-	types = {"text/html": 30,
-	         "application/xhtml+xml": 20,
-	         "text/plain": 10}
-	cs = []
-	for c in contents:
-		ctype = c.get("type")
-		if types.has_key(ctype):
-			score = types[ctype]
-		else:
-			score = 0
-		cs.append((score, c["value"]))
-	cs.sort()
-	return cs[-1][1]
-
 def encode_references(s):
 	"""Encode characters in a Unicode string using HTML references."""
 	r = StringIO()
@@ -79,15 +64,17 @@ def sanitise_html(html, baseurl, inline, config):
 	if html is None:
 		return None
 
+	html = encode_references(html)
+
 	# sgmllib handles "<br/>/" as a SHORTTAG; this workaround from
 	# feedparser.
 	html = re.sub(r'(\S)/>', r'\1 />', html)
-	html = feedparser.resolveRelativeURIs(html, baseurl)
-	p = feedparser.HTMLSanitizer()
+	html = feedparser._resolveRelativeURIs(html, baseurl, None)
+	p = feedparser._HTMLSanitizer(None)
 	p.feed(html)
 	html = p.output()
 
-	if not inline:
+	if not inline and config["blocklevelhtml"]:
 		# If we're after some block-level HTML and the HTML doesn't
 		# start with a block-level element, then insert a <p> tag
 		# before it. This still fails when the HTML contains text, then
@@ -106,36 +93,114 @@ def sanitise_html(html, baseurl, inline, config):
 		html = output[output.find("<body>") + 6
 		              : output.rfind("</body>")].strip()
 
-	return encode_references(html)
+	box = plugins.Box(html)
+	plugins.call_hook("clean_html", config, box, baseurl, inline)
+	return box.value
 
-template_re = re.compile(r'__(.*?)__')
+def select_detail(details):
+	"""Pick the preferred type of detail from a list of details. (If the
+	argument isn't a list, treat it as a list of one.)"""
+	types = {"text/html": 30,
+	         "application/xhtml+xml": 20,
+	         "text/plain": 10}
+
+	if details is None:
+		return None
+	if type(details) is not list:
+		details = [details]
+
+	ds = []
+	for detail in details:
+		ctype = detail["type"]
+		if types.has_key(ctype):
+			score = types[ctype]
+		else:
+			score = 0
+		if detail["value"] != "":
+			ds.append((score, detail))
+	ds.sort()
+
+	if len(ds) == 0:
+		return None
+	else:
+		return ds[-1][1]
+
+def detail_to_html(details, inline, config, force_preformatted = False):
+	"""Convert a detail hash or list of detail hashes as returned by
+	feedparser into HTML."""
+	detail = select_detail(details)
+	if detail is None:
+		return None
+
+	if force_preformatted:
+		html = "<pre>" + cgi.escape(detail["value"]) + "</pre>"
+	elif detail["type"] == "text/plain":
+		html = cgi.escape(detail["value"])
+	else:
+		html = detail["value"]
+
+	return sanitise_html(html, detail["base"], inline, config)
+
+def author_to_html(entry, feedurl, config):
+	"""Convert feedparser author information to HTML."""
+	author_detail = entry.get("author_detail")
+
+	if author_detail is not None and author_detail.has_key("name"):
+		name = author_detail["name"]
+	else:
+		name = entry.get("author")
+
+	url = None
+	if author_detail is not None:
+		if author_detail.has_key("url"):
+			url = author_detail["url"]
+		elif author_detail.has_key("email"):
+			url = "mailto:" + author_detail["email"]
+
+	if url is None:
+		html = name
+	else:
+		html = "<a href=\"" + cgi.escape(url) + "\">" + cgi.escape(name) + "</a>"
+
+	# We shouldn't need a base URL here anyway.
+	return sanitise_html(html, feedurl, True, config)
+
+def string_to_html(s, config):
+	"""Convert a string to HTML."""
+	return sanitise_html(cgi.escape(s), "", True, config)
+
+template_re = re.compile(r'(__.*?__)')
 def fill_template(template, bits):
 	"""Expand a template, replacing __x__ with bits["x"], and only
-	including sections bracketed by __if_x__ .. __endif__ if bits["x"]
-	is not "" (these cannot be nested). If not bits.has_key("x"),
+	including sections bracketed by __if_x__ .. [__else__ ..]
+	__endif__ if bits["x"] is not "". If not bits.has_key("x"),
 	__x__ expands to ""."""
+	result = plugins.Box()
+	plugins.call_hook("fill_template", template, bits, result)
+	if result.value is not None:
+		return result.value
+
 	f = StringIO()
-	l = template_re.split(template)
-	i = 0
-	writing = 1
-	while 1:
-		if writing:
-			f.write(l[i])
-		i += 1
-		if i == len(l):
-			break
-		key = l[i]
-		if key.startswith("if_"):
-			k = key[3:]
-			if bits.has_key(k) and bits[k] != "":
-				writing = 1
-			else:
-				writing = 0
-		elif key == "endif":
-			writing = 1
-		elif bits.has_key(key):
-			f.write(bits[key])
-		i += 1
+	if_stack = []
+	def write(s):
+		if not False in if_stack:
+			f.write(s)
+	for part in template_re.split(template):
+		if part.startswith("__") and part.endswith("__"):
+			key = part[2:-2]
+			if key.startswith("if_"):
+				k = key[3:]
+				if_stack.append(bits.has_key(k) and bits[k] != "")
+			elif key == "endif":
+				if if_stack != []:
+					if_stack.pop()
+			elif key == "else":
+				if if_stack != []:
+					if_stack.append(not if_stack.pop())
+			elif bits.has_key(key):
+				write(bits[key])
+		else:
+			write(part)
 	return f.getvalue()
 
 file_cache = {}
@@ -152,6 +217,7 @@ def short_hash(s):
 	"""Return a human-manipulatable 'short hash' of a string."""
 	return sha.new(s).hexdigest()[-8:]
 
+non_alphanumeric_re = re.compile(r'<[^>]*>|\&[^\;]*\;|[^a-z0-9]')
 class Feed:
 	"""An RSS feed."""
 
@@ -161,10 +227,8 @@ class Feed:
 		self.args = {}
 		self.etag = None
 		self.modified = None
-		self.title = None
-		self.link = None
 		self.last_update = 0
-		self.baseurl = url
+		self.feed_info = {}
 
 	def needs_update(self, now):
 		"""Return 1 if it's time to update this feed, or 0 if its
@@ -174,27 +238,46 @@ class Feed:
 		else:
 			return 1
 	
-	def update(self, articles, now, config):
+	def update(self, rawdog, now, config):
 		"""Fetch articles from a feed and add them to the collection.
-		Returns 1 if any articles were read, 0 otherwise."""
+		Returns True if any articles were read, False otherwise."""
+
+		articles = rawdog.articles
+		handlers = []
+
+		class DummyPasswordMgr:
+			def __init__(self, creds):
+				self.creds = creds
+			def add_password(self, realm, uri, user, passwd):
+				pass
+			def find_user_password(self, realm, authuri):
+				return self.creds
 
 		if self.args.has_key("user") and self.args.has_key("password"):
-			authinfo = (self.args["user"], self.args["password"])
-		else:
-			authinfo = None
+			mgr = DummyPasswordMgr((self.args["user"], self.args["password"]))
+			handlers.append(urllib2.HTTPBasicAuthHandler(mgr))
+
 		proxies = {}
 		for key in self.args.keys():
 			if key.endswith("_proxy"):
 				proxies[key[:-6]] = self.args[key]
-		if len(proxies.keys()) == 0:
-			proxies = None
+		if len(proxies.keys()) != 0:
+			handlers.append(urllib2.ProxyHandler(proxies))
 
-		feedparser.FeedParser.can_contain_relative_uris = []
-		feedparser.FeedParser.can_contain_dangerous_markup = []
+		if self.args.has_key("proxyuser") and self.args.has_key("proxypassword"):
+			mgr = DummyPasswordMgr((self.args["proxyuser"], self.args["proxypassword"]))
+			handlers.append(urllib2.ProxyBasicAuthHandler(mgr))
+
+		plugins.call_hook("add_urllib2_handlers", rawdog, config, self, handlers)
+
+		feedparser._FeedParserMixin.can_contain_relative_uris = ["url"]
+		feedparser._FeedParserMixin.can_contain_dangerous_markup = []
 		try:
-			p = feedparser.parse(self.url, self.etag,
-				self.modified,	"rawdog/" + VERSION,
-				None, authinfo, proxies)
+			p = feedparser.parse(self.url,
+				etag = self.etag,
+				modified = self.modified,
+				agent = "rawdog/" + VERSION,
+				handlers = handlers)
 			status = p.get("status")
 		except:
 			p = None
@@ -203,20 +286,29 @@ class Feed:
 		self.last_update = now
 
 		error = None
-		non_fatal = 0
+		non_fatal = False
+		old_url = self.url
 		if p is None:
 			error = "Error parsing feed."
-		elif status is None:
+		elif status is None and len(p["feed"]) == 0:
 			if config["ignoretimeouts"]:
-				return 0
+				return False
 			else:
 				error = "Timeout while reading feed."
+		elif status is None:
+			# Fetched by some protocol that doesn't have status.
+			pass
 		elif status == 301:
 			# Permanent redirect. The feed URL needs changing.
+
 			error = "New URL:     " + p["url"] + "\n"
 			error += "The feed has moved permanently to a new URL.\n"
-			error += "You should update its entry in your config file."
-			non_fatal = 1
+			if config["changeconfig"]:
+				rawdog.change_feed_url(self.url, p["url"])
+				error += "The config file has been updated automatically."
+			else:
+				error += "You should update its entry in your config file."
+			non_fatal = True
 		elif status in [403, 410]:
 			# The feed is disallowed or gone. The feed should be unsubscribed.
 			error = "The feed has gone.\n"
@@ -226,146 +318,127 @@ class Feed:
 			error = "The feed returned an error.\n"
 			error += "If this condition persists, you should remove it from your config file."
 
+		plugins.call_hook("feed_fetched", rawdog, config, self, p, error, non_fatal)
+
 		if error is not None:
-			print >>sys.stderr, "Feed:        " + self.url
+			print >>sys.stderr, "Feed:        " + old_url
 			if status is not None:
 				print >>sys.stderr, "HTTP Status: " + str(status)
 			print >>sys.stderr, error
 			print >>sys.stderr
 			if not non_fatal:
-				return 0
+				return False
 
 		self.etag = p.get("etag")
 		self.modified = p.get("modified")
 
-		self.encoding = p.get("encoding")
-		if self.encoding is None:
-			self.encoding = "utf-8"
-
-		self.baseurl = self.url
-		if p.has_key("headers") and p["headers"].has_key("content-location"):
-			self.baseurl = p["headers"]["content-location"]
-		elif p.has_key("url"):
-			self.baseurl = p["url"]
-
-		channel = p["channel"]
-		if channel.has_key("title"):
-			self.title = self.decode(channel["title"])
-		if channel.has_key("link"):
-			self.link = self.decode(channel["link"])
-
 		# In the event that the feed hasn't changed, then both channel
 		# and feed will be empty. In this case we return 0 so that
 		# we know not to expire articles that came from this feed.
-		if len(p["items"]) == 0:
-			return 0
+		if len(p["entries"]) == 0:
+			return False
 
+		self.feed_info = p["feed"]
 		feed = self.url
 
-		if config["currentonly"]:
-			for (hash, a) in articles.items():
-				if a.feed == feed:
-					del articles[hash]
-
+		seen = {}
 		sequence = 0
-		for item in p["items"]:
-			title = self.decode(item.get("title"))
-			link = self.decode(item.get("link"))
-
-			date = item.get("date_parsed")
-			if date is not None:
-				try:
-					date = time.mktime(date)
-				except OverflowError:
-					date = None
-
-			description = None
-			if description is None and item.has_key("content"):
-				description = self.decode(item["content"])
-			if description is None:
-				description = self.decode(item.get("description"))
-
-			article = Article(feed, title, link, description,
-				now, sequence, date)
+		for entry_info in p["entries"]:
+			article = Article(feed, entry_info, now, sequence)
+			ignore = plugins.Box(False)
+			plugins.call_hook("article_seen", rawdog, config, article, ignore)
+			if ignore.value:
+				continue
+			seen[article.hash] = True
 			sequence += 1
 
 			if articles.has_key(article.hash):
-				articles[article.hash].last_seen = now
+				articles[article.hash].update_from(article, now)
+				plugins.call_hook("article_updated", rawdog, config, article, now)
 			else:
 				articles[article.hash] = article
+				plugins.call_hook("article_added", rawdog, config, article, now)
 
-		return 1
+		if config["currentonly"]:
+			for (hash, a) in articles.items():
+				if a.feed == feed and not seen.has_key(hash):
+					del articles[hash]
 
-	def decode(self, s):
-		"""Convert a string or alternatives list retrieved from the
-		feed from its original encoding to our target encoding for HTML
-		output."""
-		if s is None:
-			return None
-		if type(s) != str:
-			s = select_content(s)
-		try:
-			us = s.decode(self.encoding)
-		except ValueError:
-			# Badly-encoded string (or misguessed encoding).
-			us = s
-		except LookupError:
-			# Unknown encoding.
-			us = s
-		return encode_references(us)
+		return True
 
-	def get_html_name(self):
-		if self.title is not None:
-			return self.title
-		elif self.link is not None:
-			return self.link
+	def get_html_name(self, config):
+		if self.feed_info.has_key("title_detail"):
+			r = detail_to_html(self.feed_info["title_detail"], True, config)
+		elif self.feed_info.has_key("link"):
+			r = string_to_html(self.feed_info["link"], config)
 		else:
-			return self.url
+			r = string_to_html(self.url, config)
+		if r is None:
+			r = ""
+		return r
 
 	def get_html_link(self, config):
-		s = sanitise_html(self.get_html_name(), self.get_baseurl(), 1, config)
-		if self.link is not None:
-			return '<a href="' + cgi.escape(self.link) + '">' + s + '</a>'
+		s = self.get_html_name(config)
+		if self.feed_info.has_key("link"):
+			return '<a href="' + string_to_html(self.feed_info["link"], config) + '">' + s + '</a>'
 		else:
 			return s
 
-	def get_baseurl(self):
-		try:
-			return self.baseurl
-		except AttributeError:
-			# This Feed came from an old state file.
-			return self.url
+	def get_id(self, config):
+		if self.args.has_key("id"):
+			return self.args["id"]
+		else:
+			r = self.get_html_name(config).lower()
+			return non_alphanumeric_re.sub('', r)
 
 class Article:
 	"""An article retrieved from an RSS feed."""
 
-	def __init__(self, feed, title, link, description, now, sequence, date):
+	def __init__(self, feed, entry_info, now, sequence):
 		self.feed = feed
-		self.title = title
-		self.link = link
-		self.description = description
+		self.entry_info = entry_info
 		self.sequence = sequence
-		self.date = date
 
-		s = str(feed) + str(title) + str(link) + str(description)
-		self.hash = sha.new(s).hexdigest()
+		modified = entry_info.get("modified_parsed")
+		self.date = None
+		if modified is not None:
+			try:
+				self.date = calendar.timegm(modified)
+			except OverflowError:
+				pass
+
+		self.hash = self.compute_hash()
 
 		self.last_seen = now
 		self.added = now
 
-	def get_sequence(self):
-		try:
-			return self.sequence
-		except AttributeError:
-			# This Article came from an old state file.
-			return 0
+	def compute_hash(self):
+		h = sha.new()
+		def add_hash(s):
+			h.update(s.encode("UTF-8"))
 
-	def get_date(self):
-		try:
-			return self.date
-		except AttributeError:
-			# This Article came from an old state file.
-			return None
+		add_hash(self.feed)
+		entry_info = self.entry_info
+		if entry_info.has_key("title_raw"):
+			add_hash(entry_info["title_raw"])
+		if entry_info.has_key("link"):
+			add_hash(entry_info["link"])
+		if entry_info.has_key("content"):
+			for content in entry_info["content"]:
+				add_hash(content["value_raw"])
+		if entry_info.has_key("summary_detail"):
+			add_hash(entry_info["summary_detail"]["value_raw"])
+
+		return h.hexdigest()
+
+	def update_from(self, new_article, now):
+		"""Update this article's contents from a newer article that's
+		been identified to be the same (i.e. has hashed the same, but
+		might have other changes that aren't part of the hash)."""
+		self.entry_info = new_article.entry_info
+		self.sequence = new_article.sequence
+		self.date = new_article.date
+		self.last_seen = now
 
 	def can_expire(self, now, config):
 		return ((now - self.last_seen) > config["expireage"])
@@ -431,14 +504,39 @@ def parse_bool(value):
 	else:
 		raise ValueError("Bad boolean value: " + value)
 
+def parse_list(value):
+	"""Parse a list of keywords separated by whitespace."""
+	return value.strip().split(None)
+
+def parse_feed_args(argparams, arglines):
+	"""Parse a list of feed arguments. Raise ConfigError if the syntax is invalid."""
+	args = {}
+	for a in argparams:
+		as = a.split("=", 1)
+		if len(as) != 2:
+			raise ConfigError("Bad feed argument in config: " + a)
+		args[as[0]] = as[1]
+	for a in arglines:
+		as = a.split(None, 1)
+		if len(as) != 2:
+			raise ConfigError("Bad argument line in config: " + a)
+		args[as[0]] = as[1]
+	return args
+
 class ConfigError(Exception): pass
 
 class Config:
 	"""The aggregator's configuration."""
 
 	def __init__(self):
+		self.files_loaded = []
+		self.reset()
+
+	def reset(self):
 		self.config = {
 			"feedslist" : [],
+			"feeddefaults" : {},
+			"defines" : {},
 			"outputfile" : "output.html",
 			"maxarticles" : 200,
 			"maxage" : 0,
@@ -455,50 +553,77 @@ class Config:
 			"ignoretimeouts" : 0,
 			"daysections" : 1,
 			"timesections" : 1,
+			"blocklevelhtml" : 1,
 			"tidyhtml" : 0,
 			"sortbyfeeddate" : 0,
 			"currentonly" : 0,
+			"hideduplicates" : "",
+			"newfeedperiod" : "3h",
+			"changeconfig": 0,
 			}
 
 	def __getitem__(self, key): return self.config[key]
 	def __setitem__(self, key, value): self.config[key] = value
 
-	def load(self, filename):
+	def reload(self):
+		self.log("Reloading config files")
+		self.reset()
+		for filename in self.files_loaded:
+			self.load(filename, False)
+
+	def load(self, filename, explicitly_loaded = True):
 		"""Load configuration from a config file."""
+		if explicitly_loaded:
+			self.files_loaded.append(filename)
+
+		lines = []
 		try:
 			f = open(filename, "r")
-			lines = f.readlines()
+			for line in f.xreadlines():
+				stripped = line.strip()
+				if stripped == "" or stripped[0] == "#":
+					continue
+				if line[0] in string.whitespace:
+					if lines == []:
+						raise ConfigError("First line in config cannot be an argument")
+					lines[-1][1].append(stripped)
+				else:
+					lines.append((stripped, []))
 			f.close()
 		except IOError:
 			raise ConfigError("Can't read config file: " + filename)
-		for line in lines:
-			line = line.strip()
+
+		for line, arglines in lines:
 			try:
-				self.load_line(line)
+				self.load_line(line, arglines)
 			except ValueError:
 				raise ConfigError("Bad value in config: " + line)
 
-	def load_line(self, line):
-		"""Process a configuration line."""
-
-		if line == "" or line[0] == "#":
-			return
+	def load_line(self, line, arglines):
+		"""Process a configuration directive."""
 
 		l = line.split(None, 1)
 		if len(l) != 2:
 			raise ConfigError("Bad line in config: " + line)
 
+		handled_arglines = False
 		if l[0] == "feed":
 			l = l[1].split(None)
 			if len(l) < 2:
 				raise ConfigError("Bad line in config: " + line)
-			args = {}
-			for a in l[2:]:
-				as = a.split("=", 1)
-				if len(as) != 2:
-					raise ConfigError("Bad feed argument in config: " + a)
-				args[as[0]] = as[1]
-			self["feedslist"].append((l[1], parse_time(l[0]), args))
+			self["feedslist"].append((l[1], parse_time(l[0]), parse_feed_args(l[2:], arglines)))
+			handled_arglines = True
+		elif l[0] == "feeddefaults":
+			self["feeddefaults"] = parse_feed_args(l[1].split(None), arglines)
+			handled_arglines = True
+		elif l[0] == "define":
+			l = l[1].split(None, 1)
+			if len(l) != 2:
+				raise ConfigError("Bad line in config: " + line)
+			self["defines"][l[0]] = l[1]
+		elif l[0] == "plugindirs":
+			for dir in parse_list(l[1]):
+				plugins.load_plugins(dir, self)
 		elif l[0] == "outputfile":
 			self["outputfile"] = l[1]
 		elif l[0] == "maxarticles":
@@ -531,55 +656,171 @@ class Config:
 			self["daysections"] = parse_bool(l[1])
 		elif l[0] == "timesections":
 			self["timesections"] = parse_bool(l[1])
+		elif l[0] == "blocklevelhtml":
+			self["blocklevelhtml"] = parse_bool(l[1])
 		elif l[0] == "tidyhtml":
 			self["tidyhtml"] = parse_bool(l[1])
 		elif l[0] == "sortbyfeeddate":
 			self["sortbyfeeddate"] = parse_bool(l[1])
 		elif l[0] == "currentonly":
 			self["currentonly"] = parse_bool(l[1])
+		elif l[0] == "hideduplicates":
+			self["hideduplicates"] = parse_list(l[1])
+		elif l[0] == "newfeedperiod":
+			self["newfeedperiod"] = l[1]
+		elif l[0] == "changeconfig":
+			self["changeconfig"] = parse_bool(l[1])
 		elif l[0] == "include":
-			self.load(l[1])
+			self.load(l[1], False)
+		elif plugins.call_hook("config_option_arglines", self, l[0], l[1], arglines):
+			handled_arglines = True
+		elif plugins.call_hook("config_option", self, l[0], l[1]):
+			pass
 		else:
 			raise ConfigError("Unknown config command: " + l[0])
 
+		if arglines != [] and not handled_arglines:
+			raise ConfigError("Bad argument lines in config after: " + line)
+
 	def log(self, *args):
 		"""If running in verbose mode, print a status message."""
 		if self["verbose"]:
 			print >>sys.stderr, "".join(map(str, args))
 
+	def bug(self, *args):
+		"""Report detection of a bug in rawdog."""
+		print >>sys.stderr, "Internal error detected in rawdog:"
+		print >>sys.stderr, "".join(map(str, args))
+		print >>sys.stderr, "This could be caused by a bug in rawdog itself or in a plugin."
+		print >>sys.stderr, "Please send this error message and your config file to the rawdog author."
+
+def edit_file(filename, editfunc):
+	"""Edit a file in place: for each line in the input file, call
+	editfunc(line, outputfile), then rename the output file over the input
+	file."""
+	newname = "%s.new-%d" % (filename, os.getpid())
+	oldfile = open(filename, "r")
+	newfile = open(newname, "w")
+	editfunc(oldfile, newfile)
+	newfile.close()
+	oldfile.close()
+	os.rename(newname, filename)
+
+class AddFeedEditor:
+	def __init__(self, feedline):
+		self.feedline = feedline
+	def edit(self, inputfile, outputfile):
+		d = inputfile.read()
+		outputfile.write(d)
+		if not d.endswith("\n"):
+			outputfile.write("\n")
+		outputfile.write(self.feedline)
+
+def add_feed(filename, url, config):
+	"""Try to add a feed to the config file."""
+	feeds = feedfinder.getFeeds(url)
+	if feeds == []:
+		print >>sys.stderr, "Cannot find any feeds in " + url
+	else:
+		feed = feeds[0]
+		print >>sys.stderr, "Adding feed " + feed
+		feedline = "feed %s %s\n" % (config["newfeedperiod"], feed)
+		edit_file(filename, AddFeedEditor(feedline).edit)
+
+class ChangeFeedEditor:
+	def __init__(self, oldurl, newurl):
+		self.oldurl = oldurl
+		self.newurl = newurl
+	def edit(self, inputfile, outputfile):
+		for line in inputfile.xreadlines():
+			ls = line.strip().split(None)
+			if len(ls) > 2 and ls[0] == "feed" and ls[2] == self.oldurl:
+				line = line.replace(self.oldurl, self.newurl, 1)
+			outputfile.write(line)
+
 class Rawdog(Persistable):
 	"""The aggregator itself."""
 
 	def __init__(self):
 		self.feeds = {}
 		self.articles = {}
+		self.state_version = STATE_VERSION
 
-	def list(self):
+	def check_state_version(self):
+		"""Check the version of the state file."""
+		try:
+			version = self.state_version
+		except AttributeError:
+			# rawdog 1.x didn't keep track of this.
+			version = 1
+		return version == STATE_VERSION
+
+	def change_feed_url(self, oldurl, newurl):
+		"""Change the URL of a feed."""
+
+		assert self.feeds.has_key(oldurl)
+		if self.feeds.has_key(newurl):
+			print >>sys.stderr, "Error: New feed URL is already subscribed; please remove the old one"
+			print >>sys.stderr, "from the config file by hand."
+			return
+
+		edit_file("config", ChangeFeedEditor(oldurl, newurl).edit)
+
+		feed = self.feeds[oldurl]
+		feed.url = newurl
+		del self.feeds[oldurl]
+		self.feeds[newurl] = feed
+
+		for article in self.articles.values():
+			if article.feed == oldurl:
+				article.feed = newurl
+
+		print >>sys.stderr, "Feed URL automatically changed."
+
+	def list(self, config):
 		for url in self.feeds.keys():
 			feed = self.feeds[url]
+			feed_info = feed.feed_info
 			print url
+			print "  ID:", feed.get_id(config)
 			print "  Hash:", short_hash(url)
-			print "  Title:", feed.title
-			print "  Link:", feed.link
-
-	def update(self, config, feedurl = None):
-		config.log("Starting update")
-		now = time.time()
-
-		set_socket_timeout(config["timeout"])
+			print "  Title:", feed.get_html_name(config)
+			print "  Link:", feed_info.get("link")
 
+	def sync_from_config(self, config):
 		seenfeeds = {}
 		for (url, period, args) in config["feedslist"]:
 			seenfeeds[url] = 1
 			if not self.feeds.has_key(url):
 				config.log("Adding new feed: ", url)
 				self.feeds[url] = Feed(url)
-			self.feeds[url].period = period
-			self.feeds[url].args = args
+				self.modified()
+			feed = self.feeds[url]
+			if feed.period != period:
+				config.log("Changed feed period: ", url)
+				feed.period = period
+				self.modified()
+			newargs = {}
+			newargs.update(config["feeddefaults"])
+			newargs.update(args)
+			if feed.args != newargs:
+				config.log("Changed feed options: ", url)
+				feed.args = newargs
+				self.modified()
 		for url in self.feeds.keys():
 			if not seenfeeds.has_key(url):
 				config.log("Removing feed: ", url)
 				del self.feeds[url]
+				for key, article in self.articles.items():
+					if article.feed == url:
+						del self.articles[key]
+				self.modified()
+
+	def update(self, config, feedurl = None):
+		config.log("Starting update")
+		now = time.time()
+
+		set_socket_timeout(config["timeout"])
 
 		if feedurl is None:
 			update_feeds = [url for url in self.feeds.keys()
@@ -600,15 +841,18 @@ class Rawdog(Persistable):
 		for url in update_feeds:
 			count += 1
 			config.log("Updating feed ", count, " of " , numfeeds, ": ", url)
-			if self.feeds[url].update(self.articles, now, config):
+			feed = self.feeds[url]
+			plugins.call_hook("pre_update_feed", self, config, feed)
+			rc = feed.update(self, now, config)
+			plugins.call_hook("post_update_feed", self, config, feed, rc)
+			if rc:
 				seen_some_items[url] = 1
 
 		count = 0
-		for key in self.articles.keys():
-			article = self.articles[key]
-			if ((not self.feeds.has_key(article.feed))
-			    or (seen_some_items.has_key(article.feed)
-			        and article.can_expire(now, config))):
+		for key, article in self.articles.items():
+			if (seen_some_items.has_key(article.feed)
+			    and article.can_expire(now, config)):
+				plugins.call_hook("article_expired", self, config, article, now)
 				count += 1
 				del self.articles[key]
 		config.log("Expired ", count, " articles, leaving ", len(self.articles.keys()))
@@ -625,6 +869,7 @@ class Rawdog(Persistable):
 <html lang="en">
 <head>
     <meta http-equiv="Content-Type" content="text/html; charset=ISO-8859-1">
+    <meta name="robots" content="noindex,nofollow,noarchive">
 """
 		if config["userefresh"]:
 			template += """__refresh__
@@ -660,7 +905,7 @@ by <a href="mailto:azz@us-lot.org">Adam Sampson</a>.</p>
 		if config["itemtemplate"] != "default":
 			return load_file(config["itemtemplate"])
 
-		template = """<div class="item feed-__feed_hash__" id="item-__hash__">
+		template = """<div class="item feed-__feed_hash__ feed-__feed_id__" id="item-__hash__">
 <p class="itemheader">
 <span class="itemtitle">__title__</span>
 <span class="itemfrom">[__feed_title__]</span>
@@ -685,6 +930,7 @@ __description__
 		now = time.time()
 
 		bits = { "version" : VERSION }
+		bits.update(config["defines"])
 
 		refresh = config["expireage"]
 		for feed in self.feeds.values():
@@ -696,10 +942,11 @@ __description__
 		articles = self.articles.values()
 		for a in articles:
 			if config["sortbyfeeddate"]:
-				article_dates[a] = a.get_date() or a.added
+				article_dates[a] = a.date or a.added
 			else:
 				article_dates[a] = a.added
 		numarticles = len(articles)
+
 		def compare(a, b):
 			"""Compare two articles to decide how they
 			   should be sorted. Sort by added date, then
@@ -710,45 +957,86 @@ __description__
 			i = cmp(a.feed, b.feed)
 			if i != 0:
 				return i
-			i = cmp(a.get_sequence(), b.get_sequence())
+			i = cmp(a.sequence, b.sequence)
 			if i != 0:
 				return i
 			return cmp(a.hash, b.hash)
+		plugins.call_hook("output_filter", self, config, articles)
 		articles.sort(compare)
+		plugins.call_hook("output_sort", self, config, articles)
+
 		if config["maxarticles"] != 0:
 			articles = articles[:config["maxarticles"]]
 
+		plugins.call_hook("output_write", self, config, articles)
+
 		f = StringIO()
 		itemtemplate = self.get_itemtemplate(config)
 		dw = DayWriter(f, config)
+		plugins.call_hook("output_items_begin", self, config, f)
+
+		seen_links = {}
+		seen_guids = {}
 
 		count = 0
+		dup_count = 0
 		for article in articles:
 			age = now - article.added
 			if config["maxage"] != 0 and age > config["maxage"]:
 				break
 
-			count += 1
-			dw.time(article_dates[article])
-
-			itembits = {}
-
 			feed = self.feeds[article.feed]
-			baseurl = feed.get_baseurl()
-			title = sanitise_html(article.title, baseurl, 1, config)
-			if title == "":
-				title = None
-			link = article.link
+			feed_info = feed.feed_info
+			entry_info = article.entry_info
+
+			link = entry_info.get("link")
 			if link == "":
 				link = None
-			if feed.args.has_key("format") and feed.args["format"] == "text":
-				description = "<pre>" + cgi.escape(article.description) + "</pre>"
-			else:
-				description = sanitise_html(article.description, baseurl, 0, config)
-			if description == "":
-				description = None
 
-			date = article.get_date()
+			guid = entry_info.get("id")
+			if guid == "":
+				guid = None
+
+			if feed.args.get("allowduplicates") != "true":
+				is_dup = False
+				for key in config["hideduplicates"]:
+					if key == "id" and guid is not None:
+						if seen_guids.has_key(guid):
+							is_dup = True
+						seen_guids[guid] = 1
+						break
+					elif key == "link" and link is not None:
+						if seen_links.has_key(link):
+							is_dup = True
+						seen_links[link] = 1
+						break
+				if is_dup:
+					dup_count += 1
+					continue
+
+			count += 1
+			if not plugins.call_hook("output_items_heading", self, config, f, article, article_dates[article]):
+				dw.time(article_dates[article])
+
+			itembits = {}
+			for name, value in feed.args.items():
+				if name.startswith("define_"):
+					itembits[name[7:]] = value
+
+			title = detail_to_html(entry_info.get("title_detail"), True, config)
+
+			key = None
+			for k in ["content", "summary_detail"]:
+				if entry_info.has_key(k):
+					key = k
+					break
+			if key is None:
+				description = None
+			else:
+				force_preformatted = feed.args.has_key("format") and (feed.args["format"] == "text")
+				description = detail_to_html(entry_info[key], False, config, force_preformatted)
+
+			date = article.date
 			if title is None:
 				if link is None:
 					title = "Article"
@@ -757,18 +1045,23 @@ __description__
 
 			itembits["title_no_link"] = title
 			if link is not None:
-				itembits["url"] = link
+				itembits["url"] = string_to_html(link, config)
 			else:
 				itembits["url"] = ""
+			if guid is not None:
+				itembits["guid"] = string_to_html(guid, config)
+			else:
+				itembits["guid"] = ""
 			if link is None:
 				itembits["title"] = title
 			else:
-				itembits["title"] = '<a href="' + cgi.escape(link) + '">' + title + '</a>'
+				itembits["title"] = '<a href="' + string_to_html(link, config) + '">' + title + '</a>'
 
-			itembits["feed_title_no_link"] = sanitise_html(feed.title, baseurl, 1, config)
+			itembits["feed_title_no_link"] = detail_to_html(feed_info.get("title_detail"), True, config)
 			itembits["feed_title"] = feed.get_html_link(config)
-			itembits["feed_url"] = feed.url
+			itembits["feed_url"] = string_to_html(feed.url, config)
 			itembits["feed_hash"] = short_hash(feed.url)
+			itembits["feed_id"] = feed.get_id(config)
 			itembits["hash"] = short_hash(article.hash)
 
 			if description is not None:
@@ -776,18 +1069,27 @@ __description__
 			else:
 				itembits["description"] = ""
 
+			author = author_to_html(entry_info, feed.url, config)
+			if author is not None:
+				itembits["author"] = author
+			else:
+				itembits["author"] = ""
+
 			itembits["added"] = format_time(article.added, config)
 			if date is not None:
 				itembits["date"] = format_time(date, config)
 			else:
 				itembits["date"] = ""
 
+			plugins.call_hook("output_item_bits", self, config, feed, article, itembits)
 			f.write(fill_template(itemtemplate, itembits))
 
 		dw.close()
+		plugins.call_hook("output_items_end", self, config, f)
+
 		bits["items"] = f.getvalue()
 		bits["num_items"] = str(numarticles)
-		config.log("Selected ", count, " of ", numarticles, " articles to write")
+		config.log("Selected ", count, " of ", numarticles, " articles to write; ignored ", dup_count, " duplicates")
 
 		f = StringIO()
 		print >>f, """<table id="feeds">
@@ -795,7 +1097,7 @@ __description__
 <th>Feed</th><th>RSS</th><th>Last update</th><th>Next update</th>
 </tr>"""
 		feeds = self.feeds.values()
-		feeds.sort(lambda a, b: cmp(a.get_html_name().lower(), b.get_html_name().lower()))
+		feeds.sort(lambda a, b: cmp(a.get_html_name(config).lower(), b.get_html_name(config).lower()))
 		for feed in feeds:
 			print >>f, '<tr class="feedsrow">'
 			print >>f, '<td>' + feed.get_html_link(config) + '</td>'
@@ -807,13 +1109,18 @@ __description__
 		bits["feeds"] = f.getvalue()
 		bits["num_feeds"] = str(len(feeds))
 
+		plugins.call_hook("output_bits", self, config, bits)
 		s = fill_template(self.get_template(config), bits)
 		if outputfile == "-":
 			print s
 		else:
 			config.log("Writing output file: ", outputfile)
 			f = open(outputfile + ".new", "w")
-			print >>f, s
+			try:
+				print >>f, s
+			except UnicodeEncodeError, e:
+				config.bug("Error encoding output as ASCII; UTF-8 has been written instead.\n", e)
+				print >>f, s.encode("UTF-8")
 			f.close()
 			os.rename(outputfile + ".new", outputfile)
 
@@ -837,6 +1144,12 @@ Actions (performed in order given):
 -c|--config FILE             Read additional config file FILE
 -t, --show-template          Print the template currently in use
 -T, --show-itemtemplate      Print the item template currently in use
+-a|--add URL                 Try to find a feed associated with URL and
+                             add it to the config file
+
+Special actions (all other options are ignored if one of these is specified):
+--upgrade OLDDIR NEWDIR      Import feed state from rawdog 1.x directory
+                             OLDDIR into rawdog 2.x directory NEWDIR
 
 Report bugs to <azz@us-lot.org>."""
 
@@ -844,26 +1157,31 @@ def main(argv):
 	"""The command-line interface to the aggregator."""
 
 	try:
-		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:v", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose"])
+		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add="])
 	except getopt.GetoptError, s:
 		print s
 		usage()
 		return 1
 
+	for o, a in optlist:
+		if o == "--upgrade" and len(args) == 2:
+			import upgrade_1_2
+			return upgrade_1_2.upgrade(args[0], args[1])
+
+	if len(args) != 0:
+		usage()
+		return 1
+
 	statedir = os.environ["HOME"] + "/.rawdog"
+	verbose = 0
 	for o, a in optlist:
 		if o == "--help":
 			usage()
 			return 0
 		elif o in ("-d", "--dir"):
 			statedir = a
-
-	# Support old option syntax.
-	for action in args:
-		if action in ("list", "update", "write"):
-			optlist.append(("--" + action, None))
-		else:
-			optlist.append(("--update-feed", action))
+		elif o in ("-v", "--verbose"):
+			verbose = 1
 
 	try:
 		os.chdir(statedir)
@@ -878,24 +1196,38 @@ def main(argv):
 		print >>sys.stderr, "In config:"
 		print >>sys.stderr, err
 		return 1
+	if verbose:
+		config["verbose"] = True
 
 	persister = Persister("state", Rawdog)
 	try:
 		rawdog = persister.load()
+	except KeyboardInterrupt:
+		return 1
 	except:
 		print "An error occurred while reading state from " + statedir + "/state."
 		print "This usually means the file is corrupt, and removing it will fix the problem."
 		return 1
 
+	if not rawdog.check_state_version():
+		print "The state file " + statedir + "/state was created by an older"
+		print "version of rawdog, and cannot be read by this version."
+		print "Removing the state file will fix it."
+		return 1
+
+	rawdog.sync_from_config(config)
+
+	plugins.call_hook("startup", rawdog, config)
+
 	for o, a in optlist:
 		if o in ("-u", "--update"):
 			rawdog.update(config)
-		elif o in ("-l", "--list"):
-			rawdog.list()
-		elif o in ("-w", "--write"):
-			rawdog.write(config)
 		elif o in ("-f", "--update-feed"):
 			rawdog.update(config, a)
+		elif o in ("-l", "--list"):
+			rawdog.list(config)
+		elif o in ("-w", "--write"):
+			rawdog.write(config)
 		elif o in ("-c", "--config"):
 			try:
 				config.load(a)
@@ -907,8 +1239,12 @@ def main(argv):
 			rawdog.show_template(config)
 		elif o in ("-T", "--show-itemtemplate"):
 			rawdog.show_itemtemplate(config)
-		elif o in ("-v", "--verbose"):
-			config["verbose"] = 1
+		elif o in ("-a", "--add"):
+			add_feed("config", a, config)
+			config.reload()
+			rawdog.sync_from_config(config)
+
+	plugins.call_hook("shutdown", rawdog, config)
 
 	persister.save()
 
